import vtkmodules.all as vtk
import numpy as np

##Class to handel the calculation of the mesh attributes and all color-overlays.
class Analysis():

    ##Method to calculate the area values for the triangles of a mesh.
    #@param mesh The mesh to extract the traingle area values from.
    #Returns a vtkFloatArray.
    def triangleArea(self, mesh):

        qualityFilter = vtk.vtkMeshQuality()
        qualityFilter.SetInputData(mesh)
        qualityFilter.SetTriangleQualityMeasureToArea()
        qualityFilter.Update()

        qualityArray = qualityFilter.GetOutput().GetCellData().GetArray("Quality")

        return qualityArray

    ##Method to calculate the perimeter values for the triangles of a mesh.
    #@param mesh The mesh to extract the traingle perimeter values from.
    #Returns a vtkFloatArray.
    def trainglePerimeter(self,mesh):

        result = vtk.vtkFloatArray()
        for i in range(mesh.GetNumberOfCells()):
            points = mesh.GetCell(i).GetPoints()

            point1 = points.GetPoint(0)
            point2 = points.GetPoint(1)
            point3 = points.GetPoint(2)

            edgeLength1 = np.sqrt(pow(point2[0] - point1[0],2) + pow(point2[1] - point1[1],2) + pow(point2[2] - point1[2],2))
            edgeLength2 = np.sqrt(pow(point3[0] - point2[0],2) + pow(point3[1] - point2[1],2) + pow(point3[2] - point2[2],2))
            edgeLength3 = np.sqrt(pow(point1[0] - point3[0],2) + pow(point1[1] - point3[1],2) + pow(point1[2] - point3[2],2))

            result.InsertNextValue(edgeLength1+edgeLength2+edgeLength3)

        return result

    ##Method to calculate the cange in angle for the corners of all triangles, between the original mesh and the unfolded mesh.
    #The final value is calculated as the sum of all differences of all the corresponding angles in a triangle divided by 3.
    #@param mesh The mesh to extract the traingle angle values from.
    #@param unfolded The unfolded mesh to extract the traingle angle values from.
    #Returns a vtkFloatArray.
    def triangleAngleChange(self,mesh,unfolded):
        result = vtk.vtkFloatArray()
        for i in range(mesh.GetNumberOfCells()):
            anglesMesh = self.triangleAngle(mesh.GetCell(i).GetPoints())
            anglesUnfolded = self.triangleAngle(unfolded.GetCell(i).GetPoints())
            dif = (abs(anglesMesh[0]-anglesUnfolded[0])+abs(anglesMesh[1]-anglesUnfolded[1])+abs(anglesMesh[2]-anglesUnfolded[2]))/3.0
            result.InsertNextValue(dif)
        return result

    ##Method to calculate the interior angles for all points of a triangle.
    #@param points The triangle points to calculate the angles.
    #Returns a list of length 3 containing the angles.
    def triangleAngle(self,points):

        point1 = np.array(points.GetPoint(0))
        point2 = np.array(points.GetPoint(1))
        point3 = np.array(points.GetPoint(2))

        A = point2 - point1
        B = point3 - point2
        C = point1 - point3

        angles = []
        for e1, e2 in ((A, -B), (B, -C), (C, -A)):
            num = np.dot(e1, e2)
            denom = np.linalg.norm(e1) * np.linalg.norm(e2)
            angles.append(np.arccos(num / denom) * 180 / np.pi)
        return angles

    ##Method to extract the x,y and z values for all cells of a mesh.
    #@param mesh The mesh to extract the coordinates from.
    #The returned values are calculated as the center of each cell.
    #Returns a vtkFloatArray for each axis.
    def extractXYZ(self, mesh):

        X = vtk.vtkFloatArray()
        Y = vtk.vtkFloatArray()
        Z = vtk.vtkFloatArray()

        for i in range(mesh.GetNumberOfCells()):

            x = (mesh.GetCell(i).GetPoints().GetPoint(0)[0] + mesh.GetCell(i).GetPoints().GetPoint(1)[0] + mesh.GetCell(i).GetPoints().GetPoint(2)[0]) / 3.0
            y = (mesh.GetCell(i).GetPoints().GetPoint(0)[1] + mesh.GetCell(i).GetPoints().GetPoint(1)[1] + mesh.GetCell(i).GetPoints().GetPoint(2)[1]) / 3.0
            z = (mesh.GetCell(i).GetPoints().GetPoint(0)[2] + mesh.GetCell(i).GetPoints().GetPoint(1)[2] + mesh.GetCell(i).GetPoints().GetPoint(2)[2]) / 3.0

            X.InsertNextValue(x)
            Y.InsertNextValue(y)
            Z.InsertNextValue(z)

        return X,Y,Z

    ##Helper method to calculate a difference array for two vtkFloatArrays()
    #@param old A vtkFloatArray containing the selected attribute values for all cells of the original mesh.
    #@param new A vtkFloatArray containing the selected attribute values for all cells of the unfolded mesh.
    #Returns a vtkFloatArray containg the change between the cells of the two meshes.
    def calcChange(self,old,new):
        result = vtk.vtkFloatArray()
        for i in range(old.GetNumberOfValues()):
            v = abs(old.GetValue(i) - new.GetValue(i))
            result.InsertNextValue(v)
        return result

    ##A method to create a color-overlay for a given mesh and attribute.
    #The colors will lie between perfect red and perfect blue.
    #@param geometry The geometry that is duplicated and assigned new cell color corresponding to a given vtkFloatArray of same size.
    #@param attribute The vtkFloatArray to extract the intensity of color values from.
    #Returns the color-overlay mesh.
    def generateColorMesh(self, geometry, attribute):

        newGeometry = vtk.vtkPolyData()
        newGeometry.DeepCopy(geometry)

        color = vtk.vtkColorTransferFunction()
        color.AddRGBPoint(attribute.GetRange()[0], 0.0, 0.0, 255.0)
        color.AddRGBPoint(attribute.GetRange()[1] / 3.0, 255.0, 0.0, 0.0)

        cellData = vtk.vtkUnsignedCharArray()
        cellData.SetNumberOfComponents(4)

        for i in range(geometry.GetNumberOfCells()):
            col = color.GetColor(attribute.GetValue(i))
            cellColor = [col[0], col[1], col[2], 155.0]
            cellData.InsertTuple(i, cellColor)

        newGeometry.GetCellData().SetScalars(cellData)
        return newGeometry

    ##A method to create a color-overlay for a given region of a mesh and attribute.
    #The colors will lie between perfect red and perfect blue.
    #@param geometry The geometry that is duplicated and assigned new cell color corresponding to a given vtkFloatArray of same size.
    #@param attribute The vtkFloatArray to extract the intensity of color values from.
    #@param A numpy array containg all cell ids that should be colored, the other cells are set fully transparent.
    #Returns the color-overlay mesh.
    def generateColorRegionMesh(self, geometry, attribute, regionIds):

        newGeometry = vtk.vtkPolyData()
        newGeometry.DeepCopy(geometry)

        color = vtk.vtkColorTransferFunction()
        color.AddRGBPoint(attribute.GetRange()[0], 0.0, 0.0, 255.0)
        color.AddRGBPoint(attribute.GetRange()[1] / 3.0, 255.0, 0.0, 0.0)

        cellData = vtk.vtkUnsignedCharArray()
        cellData.SetNumberOfComponents(4)

        for i in range(geometry.GetNumberOfCells()):

            if i in regionIds:
                col = color.GetColor(attribute.GetValue(i))
                cellColor = [col[0],col[1],col[2],155.0]
            else:
                cellColor = [0.0, 0.0, 0.0, 0.0]

            cellData.InsertTuple(i, cellColor)

        newGeometry.GetCellData().SetScalars(cellData)
        return newGeometry

    ##Helper method to calculate the different mesh attributes.
    #@param oldmesh the vtkPolyData containing the original mesh.
    #@param newmesh the vtkPolyData containing the unfolded mesh.
    #Returns a list containg a vtkFloatArray for every mesh attribute.
    def onProjection(self,oldmesh,newMesh):

        oldArea = self.triangleArea(oldmesh)
        newArea = self.triangleArea(newMesh)

        X,Y,Z = self.extractXYZ(oldmesh)
        areaChange = self.calcChange(oldArea,newArea)
        perimeterChange = self.calcChange(self.trainglePerimeter(oldmesh),self.trainglePerimeter(newMesh))
        angleChange = self.triangleAngleChange(oldmesh,newMesh)

        return [X,Y,Z,areaChange,perimeterChange,angleChange]



