import vtkmodules.all as vtk
import vtkmodules.numpy_interface.dataset_adapter as vtkToNp
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
import sys
from PyQt5 import QtWidgets
import os
import projection
import analysis

##The main window of the application.
#Contains all callback and visual output methods.
class Ui_MainWindow(object):

    dirname = os.path.dirname(__file__)

    ##The object to handle projections of the vertices.
    projector = projection.Projection()
    ##The object to handle the analysis of the projection.
    analyst = analysis.Analysis()
    ##The actor for the displayed imported mesh.
    mainActor = vtk.vtkActor()
    ##The actor for the unfolded mesh.
    unfoldedActor = vtk.vtkActor()
    ##The actor for the cutted mesh.
    intermediateActor = vtk.vtkActor()
    ##The actor for the color-overlay of the displayed imported mesh.
    colorActorMainMesh = vtk.vtkActor()
    ##The actor for the color-overlay of the unfolded mesh.
    colorActorUnfoldedMesh = vtk.vtkActor()
    ##The names of the unfoldData attributes.
    names = ["X", "Y", "Z", "Area", "Perimeter", "Angle"]
    ##The data of the analysis of the unfolded mesh.
    #Filled on projection.
    unfoldData = []

    ##Method to set up the Ui_Mainwindow
    #@param MainWindow
    #@param ren The vtkRenderer to display the imported mesh.
    #@param annotateRen The vtkRenderer to display color-overlay and highlighted cells.
    #@param style The vtkInteractorStyle implementing a custom left click event.
    def setupUi(self, MainWindow, ren, annotateRen, style):

        self.MainWindow = MainWindow

        self.MainWindow.setObjectName("MainWindow")

        self.style = style

        self.MainWindow.resize(1080, 720)

        self.centralWidget = QtWidgets.QWidget(MainWindow)
        self.gridlayout = QtWidgets.QGridLayout(self.centralWidget)
        self.vtkWidget = QVTKRenderWindowInteractor(self.centralWidget)
        self.gridlayout.setSpacing(2)

        groupLeft = QtWidgets.QGroupBox()
        layoutLeft = QtWidgets.QVBoxLayout()
        groupLeft.setLayout(layoutLeft)

        self.gridlayout.setColumnStretch(0, 0.5)
        self.gridlayout.setColumnStretch(1, 5)

        self.gridlayout.addWidget(groupLeft, 0, 0)
        self.gridlayout.addWidget(self.vtkWidget, 0, 1)

        self.MainWindow.setCentralWidget(self.centralWidget)

        self.setUpCamera()
        ren.SetActiveCamera(self.camera)
        annotateRen.SetActiveCamera(self.camera)

        self.unfoldedRen = vtk.vtkRenderer()
        self.unfoldedAnnotateRen = vtk.vtkRenderer()

        self.setUpUiElements(ren,annotateRen,layoutLeft)

    ##Method to create all UI elements and the associated callbacks.
    #@param ren The vtkRenderer to display the imported mesh.
    #@param annotateRen The vtkRenderer to display color-overlay and highlighted cells.
    #@layoutLeft QtBoxLayout to add buttons.
    def setUpUiElements(self,ren,annotateRen,layoutLeft):

        addFileButton = QtWidgets.QPushButton("Add .obj Files")

        ##Callback to import the main mesh.
        def getFile():
            dlg = QtWidgets.QFileDialog()
            dlg.setFileMode(QtWidgets.QFileDialog.ExistingFiles)
            dlg.setNameFilter("(*.obj)")

            filename = os.path.join(self.dirname, "./Meshes")
            dlg.setDirectory(filename)

            if dlg.exec_():
                for i in dlg.selectedFiles():
                    self.addGeometry(ren, i, dlg.directory().path())

                    name = i[:-3] + "png"

                    readerFac = vtk.vtkImageReader2Factory()
                    imageReader = readerFac.CreateImageReader2(name)
                    imageReader.SetFileName(name)

                    self.texture = vtk.vtkTexture()
                    self.texture.SetInputConnection(imageReader.GetOutputPort())
            self.vtkWidget.GetRenderWindow().Render()

        addFileButton.clicked.connect(getFile)

        cylindricalProjectionButton = QtWidgets.QPushButton("Cylindrical Projection")

        ##Callback to unfold the imported main mesh by applying a cylindrical projection.
        #Opens a message window if triggered without an imported main mesh.
        def cylindricalProjection():

            try:
                if hasattr(self,"view"):
                    self.view.GetRenderWindow().Finalize()
                    self.view.GetInteractor().TerminateApp()

                poly = self.projector.inflate(self.mainActor.GetMapper().GetInput())

                poly = self.projector.cutGeometry(poly)

                intMapper = vtk.vtkPolyDataMapper()
                intMapper.SetInputData(self.projector.cutGeometry(self.mainActor.GetMapper().GetInput()))
                self.intermediateActor.SetMapper(intMapper)
                MouseInteractorHighlightCell.setActor(self.style, self.intermediateActor)

                altPoly = self.projector.cylindricalProjection(poly)

                self.unfoldData = self.analyst.onProjection(poly, altPoly)

                mapper = vtk.vtkPolyDataMapper()
                mapper.SetInputData(altPoly)
                self.unfoldedActor.SetMapper(mapper)
                self.unfoldedActor.SetTexture(self.texture)
                self.unfoldedActor.GetProperty().BackfaceCullingOn()

                self.style.setUnfoldedActor(self.unfoldedActor)
                self.style.setUnfoldedData(self.unfoldData)
                self.style.setUnfoldedAnnotateRenderer(self.unfoldedAnnotateRen)

                ren.SetViewport([0.0, 0.0, 1.0, 0.5])
                annotateRen.SetViewport([0.0, 0.0, 1.0, 0.5])

                camera = vtk.vtkCamera()
                camera.SetViewUp(0, 0, 1)
                camera.SetPosition(0, -8, 0)
                camera.SetFocalPoint(0, 0, 0)

                self.setUpUnfoldedRenderer(camera)

                self.updateActorColors(annotateRen,self.unfoldedAnnotateRen,3 + selectedCheckBox())

                self.vtkWidget.GetRenderWindow().Render()

                self.pcp = self.parallelCoordinatePlot(self.unfoldData,self.names, annotateRen, self.unfoldedAnnotateRen)
            except:
                msg = QtWidgets.QMessageBox()
                msg.setText("Error")
                msg.setInformativeText("Please select a .obj first.")
                msg.exec_()

        cylindricalProjectionButton.clicked.connect(cylindricalProjection)

        sphericalProjectionButton = QtWidgets.QPushButton("Spherical Projection")

        ##Callback to unfold the imported main mesh by applying a spherical projection.
        #Opens a message window if triggered without an imported main mesh.
        def sphericalProjection():
            try:
                if hasattr(self,"view"):
                    self.view.GetRenderWindow().Finalize()
                    self.view.GetInteractor().TerminateApp()

                poly = self.projector.inflate(self.mainActor.GetMapper().GetInput())

                poly = self.projector.cutGeometry(poly)

                intMapper = vtk.vtkPolyDataMapper()
                intMapper.SetInputData(self.projector.cutGeometry(self.mainActor.GetMapper().GetInput()))
                self.intermediateActor.SetMapper(intMapper)
                MouseInteractorHighlightCell.setActor(self.style, self.intermediateActor)

                altPoly = self.projector.sphericalProjection(poly)

                self.unfoldData = self.analyst.onProjection(poly, altPoly)

                mapper = vtk.vtkPolyDataMapper()
                mapper.SetInputData(altPoly)

                self.unfoldedActor.SetMapper(mapper)
                self.unfoldedActor.SetTexture(self.texture)
                self.unfoldedActor.GetProperty().BackfaceCullingOn()

                self.style.setUnfoldedActor(self.unfoldedActor)
                self.style.setUnfoldedData(self.unfoldData)
                self.style.setUnfoldedAnnotateRenderer(self.unfoldedAnnotateRen)

                ren.SetViewport([0.0, 0.0, 1.0, 0.5])
                annotateRen.SetViewport([0.0, 0.0, 1.0, 0.5])

                camera = vtk.vtkCamera()
                camera.SetViewUp(0, 0, 1)
                camera.SetPosition(0, 7, 0)
                camera.SetFocalPoint(0, 0, 0)

                self.setUpUnfoldedRenderer(camera)

                self.updateActorColors(annotateRen,self.unfoldedAnnotateRen,3 + selectedCheckBox())

                self.vtkWidget.GetRenderWindow().Render()

                self.pcp = self.parallelCoordinatePlot(self.unfoldData,self.names, annotateRen,self.unfoldedAnnotateRen)
            except:
                msg = QtWidgets.QMessageBox()
                msg.setText("Error")
                msg.setInformativeText("Please select a .obj first.")
                msg.exec_()

        sphericalProjectionButton.clicked.connect(sphericalProjection)

        cleanButton = QtWidgets.QPushButton("Clear Overlay")

        ##Callback to remove all color-overlay, highlighted cells and radialplots.
        def cleanUp():
            annotateRen.RemoveAllViewProps()
            self.unfoldedAnnotateRen.RemoveAllViewProps()

        cleanButton.clicked.connect(cleanUp)

        area = QtWidgets.QPushButton("Area")
        area.setCheckable(True)
        perimeter = QtWidgets.QPushButton("Perimeter")
        perimeter.setCheckable(True)
        angle = QtWidgets.QPushButton("Angle")
        angle.setCheckable(True)

        area.setChecked(True)
        self.selectedAttribute = 0

        ##Helper method to return a integer representing the checked mesh parameter for color-overlay.
        def selectedCheckBox():
            if area.isChecked(): return 0
            elif perimeter.isChecked(): return 1
            else: return 2

        ##Callback if the area mesh parameter is clicked.
        #Color-overlay will represent the change in area per traingle between the original and the unfolded mesh.
        def onArea():
            if(self.unfoldedActor.GetMapper() != None):
                self.selectedAttribute = 0
                self.updateActorColors(annotateRen,self.unfoldedAnnotateRen,3 + self.selectedAttribute)
                self.vtkWidget.GetRenderWindow().Render()
                self.selectedAttribute = 0
            perimeter.setChecked(False)
            angle.setChecked(False)

        ##Callback if the perimeter mesh parameter is clicked.
        #Color-overlay will represent the change in perimeter per traingle between the original and the unfolded mesh.
        def onPerimeter():
            if(self.unfoldedActor.GetMapper() != None):
                self.selectedAttribute = 1
                self.updateActorColors(annotateRen,self.unfoldedAnnotateRen,3 + self.selectedAttribute)
                self.vtkWidget.GetRenderWindow().Render()
            area.setChecked(False)
            angle.setChecked(False)

        ##Callback if the angle mesh parameter is clicked.
        #Color-overlay will represent the average absolute change of angle per corner, per traingle between the original and the unfolded mesh.
        def onAngle():
            if(self.unfoldedActor.GetMapper() != None):
                self.selectedAttribute = 2
                self.updateActorColors(annotateRen,self.unfoldedAnnotateRen,3 + self.selectedAttribute)
                self.vtkWidget.GetRenderWindow().Render()
            area.setChecked(False)
            perimeter.setChecked(False)

        area.clicked.connect(onArea)
        perimeter.clicked.connect(onPerimeter)
        angle.clicked.connect(onAngle)

        generalGroup = QtWidgets.QGroupBox()
        generalLayout = QtWidgets.QVBoxLayout()
        generalGroup.setLayout(generalLayout)
        generalLayout.addWidget(addFileButton)
        generalLayout.addWidget(cleanButton)

        projectionGroup = QtWidgets.QGroupBox()
        projectionLayout = QtWidgets.QVBoxLayout()
        projectionGroup.setLayout(projectionLayout)
        projectionLayout.addWidget(sphericalProjectionButton)
        projectionLayout.addWidget(cylindricalProjectionButton)

        attributesGroup = QtWidgets.QGroupBox()
        attributesLayout = QtWidgets.QVBoxLayout()
        attributesGroup.setLayout(attributesLayout)
        attributesLayout.addWidget(area)
        attributesLayout.addWidget(perimeter)
        attributesLayout.addWidget(angle)

        layoutLeft.addWidget(generalGroup)
        layoutLeft.addWidget(projectionGroup)
        layoutLeft.addWidget(attributesGroup)

    ##Helper method to set up the vtkCamera of the renderer of the main imported mesh.
    def setUpCamera(self):
        self.camera = vtk.vtkCamera()
        self.camera.SetViewUp(0,0,1)
        self.camera.SetPosition(0, -10, 0)
        self.camera.SetFocalPoint(0, 0, 0)

    ##Helper method to create a renderer for the unfolded mesh and a renderer for the color overlay of the unfolded mesh.
    #@param camera The vtkCamera to use for both renderes
    def setUpUnfoldedRenderer(self,camera):

        self.unfoldedRen.SetLayer(0)
        self.unfoldedAnnotateRen.SetLayer(1)

        self.unfoldedRen.SetBackground(255.0, 255.0, 255.0)
        self.unfoldedRen.SetActiveCamera(camera)
        self.unfoldedRen.AddActor(self.unfoldedActor)
        self.unfoldedRen.InteractiveOff()
        self.unfoldedRen.SetViewport([0.0, 0.5, 1.0, 1.0])

        self.unfoldedAnnotateRen.SetBackground(255.0, 255.0, 255.0)
        self.unfoldedAnnotateRen.SetActiveCamera(camera)
        self.unfoldedAnnotateRen.AddActor(self.colorActorUnfoldedMesh)
        self.unfoldedAnnotateRen.InteractiveOff()
        self.unfoldedAnnotateRen.SetViewport([0.0, 0.5, 1.0, 1.0])

        self.vtkWidget.GetRenderWindow().AddRenderer(self.unfoldedRen)
        self.vtkWidget.GetRenderWindow().AddRenderer(self.unfoldedAnnotateRen)

    ##Method to add geometry to the main renderer.
    #Only used to import the main mesh.
    #@param ren The main vtkRenderer.
    #@param geometryFileName The string filename of the selected .obj file.
    #@param directory The string path of the directory.
    def addGeometry(self,ren,geometryFileName, directory):

        importer = vtk.vtkOBJImporter()
        filename = geometryFileName
        importer.SetFileName(filename)
        filename = geometryFileName[:-3] + "mtl"
        importer.SetFileNameMTL(filename)
        filename = os.path.join(directory)
        importer.SetTexturePath(filename)
        importer.SetRenderWindow(ren.GetRenderWindow())
        importer.Update()

        actors = ren.GetActors()

        actor = actors.GetLastActor()

        transform = vtk.vtkTransform()
        transform.RotateX(90.0)
        rotate = vtk.vtkTransformFilter()
        rotate.SetInputData(actor.GetMapper().GetInput())
        rotate.SetTransform(transform)
        rotate.Update()

        actor.GetMapper().SetInputConnection(rotate.GetOutputPort())
        actor.Modified()

        self.mainActor = actor
        ren.RemoveAllViewProps()
        ren.AddActor(self.mainActor)

    ##Method to create the parallelcoordinates plot and open it in an extern window.
    #Brushing callback adding color-overlay for all selected cells with the according color.
    #@param data The unfoldData list element.
    #@param dataNames A list containing the associated unfoldData labels.
    #@param ren The vtkRenderer for color-overlays, cell highlighting of the main mesh and radial plot.
    #@param unfoldedRen The vtkRenderer for color-overlays of the unfolded mesh.
    #@param selectedAttribute An integer representing the vtkFloatArray in unfoldData to use for the color-overlay.
    def parallelCoordinatePlot(self, data, dataNames, ren, unfoldedRen):

        table = vtk.vtkTable()

        polyd = vtk.vtkPolyData()

        count = 0
        for i in data:
            i.SetName(dataNames[count])
            table.AddColumn(i)
            polyd.GetPointData().AddArray(i)
            count += 1

        rep = vtk.vtkParallelCoordinatesRepresentation()
        rep.SetInputData(polyd)

        for i in range(len(data)):
            rep.SetInputArrayToProcess(i, 0, 0, 0, table.GetColumn(i).GetName())

        rep.SetUseCurves(0)
        rep.SetLineOpacity(0.1)
        rep.SetFontSize(0.6)

        annotationLink = vtk.vtkAnnotationLink()
        annotationLink.GetCurrentSelection().GetNode(0).SetFieldType(1)
        annotationLink.GetCurrentSelection().GetNode(0).SetContentType(4)
        annotationLink.Update()
        rep.SetAnnotationLink(annotationLink)

        self.view = vtk.vtkParallelCoordinatesView()
        self.view.SetRepresentation(rep)
        self.view.SetInspectMode(1)

        self.view.SetBrushModeToLasso()
        self.view.SetBrushOperatorToReplace()

        def update_render_windows(obj, event):
            self.vtkWidget.GetRenderWindow().Render()

        ##Callback to enable brushing.
        #Adds a region color-overlay over the brushed triangles.
        def selection(obj,event):
            idArr = annotationLink.GetCurrentSelection().GetNode(0).GetSelectionList()

            main = self.analyst.generateColorRegionMesh(self.intermediateActor.GetMapper().GetInput(), self.unfoldData[self.selectedAttribute+3], vtkToNp.numpy_support.vtk_to_numpy(idArr))
            unfolded = self.analyst.generateColorRegionMesh(self.unfoldedActor.GetMapper().GetInput(), self.unfoldData[self.selectedAttribute+3], vtkToNp.numpy_support.vtk_to_numpy(idArr))
            ren.RemoveAllViewProps()

            ren.RemoveAllViewProps()
            unfoldedRen.RemoveAllViewProps()

            mapperColor = vtk.vtkPolyDataMapper()
            mapperColor.SetInputData(main)
            self.colorActorMainMesh.SetMapper(mapperColor)
            self.colorActorMainMesh.GetProperty().SetOpacity(0.5)
            self.colorActorMainMesh.GetProperty().BackfaceCullingOn()
            ren.AddActor(self.colorActorMainMesh)

            mapperColor2 = vtk.vtkPolyDataMapper()
            mapperColor2.SetInputData(unfolded)
            self.colorActorUnfoldedMesh.SetMapper(mapperColor2)
            self.colorActorUnfoldedMesh.GetProperty().SetOpacity(0.5)
            self.colorActorUnfoldedMesh.GetProperty().BackfaceCullingOn()
            unfoldedRen.AddActor(self.colorActorUnfoldedMesh)

            self.spiderPlot(vtkToNp.numpy_support.vtk_to_numpy(idArr), ren)

            self.vtkWidget.GetRenderWindow().Render()


        self.view.GetInteractor().AddObserver("UserEvent", update_render_windows)
        annotationLink.AddObserver("AnnotationChangedEvent", selection)

        screenSize = self.view.GetRenderWindow().GetScreenSize()

        #self.view.GetRenderWindow().SetSize(int(screenSize[0]*0.5),int(screenSize[1]*0.25))
        #self.view.GetRenderWindow().SetPosition(int(screenSize[0]*0.25),int(screenSize[0]))

        self.view.Render()
        self.view.GetInteractor().Start()

    ##Method to create a radial plot and add it to the provided renderer.
    #@param data A numpy array containig all the indices of the cells that are selected.
    #@param ren The vtkRenderer to add the plot to.
    def spiderPlot(self, data, ren):

        if len(data) > 0:
            numTuples = 1

            areaMean = 0.0
            perimeterMean = 0.0
            angleMean = 0.0
            for i in data:
                areaMean += self.unfoldData[3].GetValue(i)
                perimeterMean += self.unfoldData[4].GetValue(i)
                angleMean += self.unfoldData[5].GetValue(i)
            areaMean /= len(data)
            perimeterMean /= len(data)
            angleMean /= len(data)

            area = vtk.vtkFloatArray()
            area.SetNumberOfTuples(numTuples)
            area.InsertTuple1(0,areaMean)

            perimeter = vtk.vtkFloatArray()
            perimeter.SetNumberOfTuples(numTuples)
            perimeter.InsertTuple1(0,perimeterMean)

            angle = vtk.vtkFloatArray()
            angle.SetNumberOfTuples(numTuples)
            angle.InsertTuple1(0,angleMean)

            dobj = vtk.vtkDataObject()
            dobj.GetFieldData().AddArray(area)
            dobj.GetFieldData().AddArray(perimeter)
            dobj.GetFieldData().AddArray(angle)

            actor = vtk.vtkSpiderPlotActor()
            actor.SetInputData(dobj)
            actor.SetIndependentVariablesToColumns()
            actor.GetPositionCoordinate().SetValue(0.7, 0.2, 0.0)
            actor.GetPosition2Coordinate().SetValue(0.95, 0.9, 0.0)
            actor.GetProperty().SetColor([0.3,0.3,0.3])

            actor.SetAxisLabel(0, "Area")
            actor.SetAxisRange(0, 0.0, self.unfoldData[3].GetRange()[1])

            actor.SetAxisLabel(1, "Perimeter")
            actor.SetAxisRange(1, 0.0, self.unfoldData[4].GetRange()[1])

            actor.SetAxisLabel(2, "Angle")
            actor.SetAxisRange(2, 0.0, self.unfoldData[5].GetRange()[1])

            actor.GetLegendActor().SetNumberOfEntries(numTuples)

            actor.SetPlotColor(0, 1.0, 0.0, 0.0)

            actor.LegendVisibilityOff()

            actor.GetLabelTextProperty().SetColor([0.0,0.0,0.0])

            ren.AddActor(actor)

    ##Helper method to update the color-overlay for both the main mesh and the unfolded mesh.
    #@param ren The vtkRenderer for the main mesh.
    #@param unfoldedRen The vtkRenderer for the unfolded mesh.
    #@param idx An integer representing the selected mesh attribute to highlight, in the unfoldData list.
    def updateActorColors(self,ren,unfoldedRen,idx):

        main = self.analyst.generateColorMesh(self.intermediateActor.GetMapper().GetInput(), self.unfoldData[idx])
        unfolded = self.analyst.generateColorMesh(self.unfoldedActor.GetMapper().GetInput(), self.unfoldData[idx])

        ren.RemoveAllViewProps()
        unfoldedRen.RemoveAllViewProps()

        mapperColor = vtk.vtkPolyDataMapper()
        mapperColor.SetInputData(main)
        self.colorActorMainMesh.SetMapper(mapperColor)
        self.colorActorMainMesh.GetProperty().SetOpacity(0.5)
        self.colorActorMainMesh.GetProperty().BackfaceCullingOn()
        ren.AddActor(self.colorActorMainMesh)

        mapperColor2 = vtk.vtkPolyDataMapper()
        mapperColor2.SetInputData(unfolded)
        self.colorActorUnfoldedMesh.SetMapper(mapperColor2)
        self.colorActorUnfoldedMesh.GetProperty().SetOpacity(0.5)
        self.colorActorUnfoldedMesh.GetProperty().BackfaceCullingOn()
        unfoldedRen.AddActor(self.colorActorUnfoldedMesh)

##Class implementing the vtkInteractorStyleTrackballCamera
#Adds a custom callback to highlight cells on a right click and draw a radial plot.
class MouseInteractorHighlightCell(vtk.vtkInteractorStyleTrackballCamera):

    cellIds = []

    def __init__(self, parent=None):
        self.AddObserver("RightButtonPressEvent", self.rightButtonPressEvent)

        self.LastPickedCell = None
        self.LastPickedProperty = vtk.vtkProperty()
        
    def setActor(self,actor):
        self.actor = actor

    def setUnfoldedActor(self,actor):
        self.unfoldedActor = actor

    def setUnfoldedData(self,data):
        self.unfoldData = data

    def setAnnotateRenderer(self,ren):
        self.ren = ren

    def setUnfoldedAnnotateRenderer(self, unfoldedRen):
        self.unfoldedRen = unfoldedRen

    def spiderPlot(self, data,color):

        numTuples = 1

        areaMean = 0.0
        perimeterMean = 0.0
        angleMean = 0.0

        areaMean += self.unfoldData[3].GetValue(data)
        perimeterMean += self.unfoldData[4].GetValue(data)
        angleMean += self.unfoldData[5].GetValue(data)

        area = vtk.vtkFloatArray()
        area.SetNumberOfTuples(numTuples)
        area.InsertTuple1(0, areaMean)

        perimeter = vtk.vtkFloatArray()
        perimeter.SetNumberOfTuples(numTuples)
        perimeter.InsertTuple1(0, perimeterMean)

        angle = vtk.vtkFloatArray()
        angle.SetNumberOfTuples(numTuples)
        angle.InsertTuple1(0, angleMean)

        dobj = vtk.vtkDataObject()
        dobj.GetFieldData().AddArray(area)
        dobj.GetFieldData().AddArray(perimeter)
        dobj.GetFieldData().AddArray(angle)

        actor = vtk.vtkSpiderPlotActor()
        actor.SetInputData(dobj)
        # actor.SetTitle("Spider Plot")
        actor.SetIndependentVariablesToColumns()
        actor.GetPositionCoordinate().SetValue(0.7, 0.2, 0.0)
        actor.GetPosition2Coordinate().SetValue(0.95, 0.9, 0.0)
        actor.GetProperty().SetColor([0.3, 0.3, 0.3])

        actor.SetAxisLabel(0, "Area")
        actor.SetAxisRange(0, 0.0, self.unfoldData[3].GetRange()[1])

        actor.SetAxisLabel(1, "Perimeter")
        actor.SetAxisRange(1, 0.0, self.unfoldData[4].GetRange()[1])

        actor.SetAxisLabel(2, "Angle")
        actor.SetAxisRange(2, 0.0, self.unfoldData[5].GetRange()[1])

        actor.GetLegendActor().SetNumberOfEntries(numTuples)

        actor.SetPlotColor(0, color)

        actor.LegendVisibilityOff()

        actor.GetLabelTextProperty().SetColor([0.0, 0.0, 0.0])

        self.ren.AddActor(actor)

    def rightButtonPressEvent(self, obj, event):
        clickPos = self.GetInteractor().GetEventPosition()

        picker = vtk.vtkCellPicker()
        picker.SetPickTextureData(False)
        picker.Pick(clickPos[0], clickPos[1], 0, self.ren)

        self.NewPickedCell = picker.GetCellId()

        if self.NewPickedCell != -1:
            color = [vtk.vtkMath.Random(), vtk.vtkMath.Random(), vtk.vtkMath.Random()]
            actorMain = self.extractTriangle(self.actor.GetMapper().GetInput(),color)
            actorUnfolded = self.extractTriangle(self.unfoldedActor.GetMapper().GetInput(),color)

            self.ren.AddActor(actorMain)
            self.unfoldedRen.AddActor(actorUnfolded)

            self.spiderPlot(self.NewPickedCell,color)

        self.OnRightButtonDown()
        return

    ##Helper method creating a vtkPolydata containing a single vtkTriangle.
    #@param mesh The original vtkPolydata to extract the vtkTriangle from.
    #@param color The color of the returned vtkActor.
    #Returns a vtkActor.
    def extractTriangle(self,mesh,color):

        points = mesh.GetCell(self.NewPickedCell).GetPoints()
        newPoints = vtk.vtkPoints()
        vertices = vtk.vtkCellArray()

        id = newPoints.InsertNextPoint(points.GetPoint(0))
        vertices.InsertNextCell(1)
        vertices.InsertCellPoint(id)

        id = newPoints.InsertNextPoint(points.GetPoint(1))
        vertices.InsertNextCell(1)
        vertices.InsertCellPoint(id)

        id = newPoints.InsertNextPoint(points.GetPoint(2))
        vertices.InsertNextCell(1)
        vertices.InsertCellPoint(id)

        triangle = vtk.vtkTriangle()
        triangle.GetPointIds().SetId(0, 0)
        triangle.GetPointIds().SetId(1, 1)
        triangle.GetPointIds().SetId(2, 2)

        triangles = vtk.vtkCellArray()
        triangles.InsertNextCell(triangle)

        poly = vtk.vtkPolyData()
        poly.SetPoints(newPoints)
        poly.SetPolys(triangles)

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputData(poly)

        actor = vtk.vtkActor()
        actor.SetMapper(mapper)

        actor.GetProperty().SetColor(color)
        actor.GetProperty().SetOpacity(0.5)
        actor.GetProperty().BackfaceCullingOn()
        return actor

##Class implementing QMainWindow to help with Qt and VTK interaction.
class SimpleView(QtWidgets.QMainWindow):

    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        self.ren = vtk.vtkRenderer()
        self.ren.SetLayer(0)
        self.annotateRen = vtk.vtkRenderer()
        self.annotateRen.SetLayer(1)
        self.ren.SetBackground(255.0,255.0,255.0)
        self.annotateRen.SetBackground(255.0,255.0,255.0)
        style = MouseInteractorHighlightCell()
        style.SetDefaultRenderer(self.ren)
        style.setAnnotateRenderer(self.annotateRen)

        self.ui = Ui_MainWindow()
        self.ui.setupUi(self,self.ren,self.annotateRen,style)
        self.ui.vtkWidget.GetRenderWindow().SetNumberOfLayers(2)
        self.ui.vtkWidget.GetRenderWindow().AddRenderer(self.ren)
        self.ui.vtkWidget.GetRenderWindow().AddRenderer(self.annotateRen)

        self.iren = self.ui.vtkWidget.GetRenderWindow().GetInteractor()
        self.iren.SetInteractorStyle(style)

        self.ui.vtkWidget.GetRenderWindow().SetMultiSamples(0)
        self.ui.vtkWidget.GetRenderWindow().SetAlphaBitPlanes(True)

if __name__ == '__main__':

    app = QtWidgets.QApplication(sys.argv)
    window = SimpleView()
    window.show()
    window.iren.Initialize()
    sys.exit(app.exec_())