import vtkmodules.all as vtk
import numpy as np

#Class to handle the inflating, cutting and unfolding of the meshes.
class Projection():

    ##Method to offset/inflate all the vertices of mesh to the surface of a sphere with default radius 2.0.
    #@param geometry The vtkPolyData of the original mesh to offset.
    #@param the radius of the sphere.
    #Returns the inflatted vtkPolyData.
    def inflate(self, geometry, r=2):

        newGeometry = vtk.vtkPolyData()
        newGeometry.DeepCopy(geometry)
        numberOfPoints = newGeometry.GetNumberOfPoints()

        points = newGeometry.GetPoints()

        for i in range(numberOfPoints):
            point = points.GetPoint(i)
            polarCoord = self.convertCartesianToPolar(point)
            x = r * np.sin((polarCoord[1])) * np.cos((polarCoord[0]))
            y = r * np.sin((polarCoord[1])) * np.sin((polarCoord[0]))
            z = r * np.cos((polarCoord[1]))
            point = [x,y,z]
            points.SetPoint(i,point)

        newGeometry.SetPoints(points)
        return newGeometry

    ##Method to project all the vertices of a given mesh to their respective longitude and latitude in polar coordinates.
    #@param geometry The vtkPolyData to map/unfold.
    #Returns a vtkPolyData of the unfolded input mesh.
    def sphericalProjection(self, geometry):

        newGeometry = vtk.vtkPolyData()
        newGeometry.DeepCopy(geometry)
        numberOfPoints = newGeometry.GetNumberOfPoints()

        points = newGeometry.GetPoints()

        for i in range(numberOfPoints):
            point = points.GetPoint(i)
            polarCoord = self.convertCartesianToPolar(point)

            divi = -1.0
            point = [((polarCoord[0])/divi),0.0,(((polarCoord[1] - np.pi / 2.0 + 0.1)/divi))]
            points.SetPoint(i, point)

        newGeometry.SetPoints(points)

        normalsGen = vtk.vtkPolyDataNormals()
        normalsGen.SetInputData(newGeometry)
        normalsGen.ComputePointNormalsOn()
        normalsGen.ComputeCellNormalsOff()
        normalsGen.Update()
        return(normalsGen.GetOutput())

    ##Method to project all the vertices of a given mesh to their respective longitude and latitude in cylindrical coordinates.
    #@param geometry The vtkPolyData to map/unfold.
    #Returns a vtkPolyData of the unfolded input mesh.
    def cylindricalProjection(self, geometry):

        newGeometry = vtk.vtkPolyData()
        newGeometry.DeepCopy(geometry)
        numberOfPoints = newGeometry.GetNumberOfPoints()

        points = newGeometry.GetPoints()

        for i in range(numberOfPoints):
            point = points.GetPoint(i)
            polarCoord = self.convertCartesionToCylindrical(point)
            divi = 1.0
            point = [((polarCoord[0]) / divi), 0.0, (((polarCoord[1]) / divi))]
            points.SetPoint(i, point)

        newGeometry.SetPoints(points)
        normalsGen = vtk.vtkPolyDataNormals()
        normalsGen.SetInputData(newGeometry)
        normalsGen.ComputePointNormalsOn()
        normalsGen.ComputeCellNormalsOff()
        normalsGen.Update()
        return(normalsGen.GetOutput())

    ##Method to convert cartesian coordinates into cylindrical coordinates.
    #param coord A vtkPoint that should be converted to cylindrical coordinates.
    def convertCartesionToCylindrical(self, coord):
        coord = np.array(coord)
        r = np.sqrt(pow(coord[0],2)+pow(coord[1],2))
        phi = np.arctan2(coord[1],coord[0])
        z = coord[2]

        return [phi,z]

    ##Method to convert cartesian coordinates into polar coordinates.
    #param coord A vtkPoint that should be converted to polar coordinates.
    def convertCartesianToPolar(self, coord):
        coord = np.array(coord)
        radius = np.sqrt((coord[0] * coord[0]) + (coord[1] * coord[1]) + (coord[2] * coord[2]))
        #long = math.acos(coord[0] / math.sqrt(coord[0] * coord[0] + coord[1] * coord[1])) * (np.sign(coord[1]))
        lat = np.arctan2(coord[1],coord[0])
        long = np.arccos(coord[2] / radius)

        return [(lat), (long)]

    ##Method to cut a given mesh along a plane with normal pointing along the positive z axis.
    #The implementation of the used vtkClipPolyData results in artifacts in the textur information, looking like cracks in the mesh.
    #@param geometry A vtkPolyData to cut.
    #Returns the cutted mesh.
    def cutGeometry(self,geometry):

        plane = vtk.vtkPlane()
        plane.SetNormal(0.0, 1.0, 0.0)
        plane.SetOrigin(0.0, 0.0, 0.0)

        clipper2 = vtk.vtkClipPolyData()
        clipper2.SetClipFunction(plane)
        clipper2.SetInputData(geometry)
        clipper2.InsideOutOn()
        clipper2.GenerateClippedOutputOn()
        clipper2.Update()

        trans = vtk.vtkTransform()
        trans.Translate(0.0,-0.00001,0.0)
        trans2 = vtk.vtkTransform()
        trans2.Translate(0.0,0.00001,0.0)

        transform = vtk.vtkTransformFilter()
        transform.SetTransform(trans)
        transform.SetInputConnection(clipper2.GetOutputPort())
        transform.Update()
        transform2 = vtk.vtkTransformFilter()
        transform2.SetTransform(trans2)
        transform2.SetInputConnection(clipper2.GetOutputPort(1))
        transform2.Update()

        append = vtk.vtkAppendPolyData()
        append.AddInputConnection(transform.GetOutputPort())
        append.AddInputConnection(transform2.GetOutputPort())
        append.Update()

        return append.GetOutput()